#include "grid.h"

#include <algorithm>
#include <math.h>
#include <random>
#include <unordered_set>
#include <unordered_map>
#include <queue>

#include <SFML/Graphics/RenderWindow.hpp>

#include "person.h"

Grid::Grid(const sf::Image& map, int& robotX, int& robotY) {
    sf::Color px;
    for (int x = 0; x < WIDTH; ++x) {
        for (int y = 0; y < HEIGHT; ++y) {
            #ifdef DRAW
            shape[x][y].setSize(sf::Vector2f(SIZE, SIZE));
            shape[x][y].setPosition(x * SIZE, y * SIZE);
            #endif

            if (x == 0 || y == 0 || x == WIDTH - 1 || y == HEIGHT - 1) {
                setWall(x, y);
                continue;
            }

            px = map.getPixel(x * map.getSize().x / WIDTH, y * map.getSize().y / HEIGHT);
            if (px == sf::Color::Green) {
                visible[x][y] = true;
                entrance.push_back(std::make_pair(x, y));
            } else if (px == sf::Color::Blue) {
                robotX = x;
                robotY = y;
            } else if (px == sf::Color::Yellow) {
                makeCamera(x, y);
                exit.push_back(std::make_pair(x, y));
            } else if (px != sf::Color::White) {
                if (px == sf::Color(245, 121, 33))
                    makeCamera(x, y);
                setWall(x, y);
            }
            #ifdef DRAW
            setColor(x, y);
            #endif
        }
    }
}

void Grid::makeCamera(const int& x, const int& y) {
    return;
    for (int i = -5; i <= 5; ++i)
        for (int j = -5; j <= 5; ++j)
            if ((i != -5 && i != 5) || (j != -5 && j != 5))
                setVisible(x + i, y + j);
}

void Grid::step() {
    if (rand() % 20 == 0) { // On average, someone appears once every 10 time units
        std::pair<int, int> pos = entrance[rand() % entrance.size()];
        people.push_back(new Person(*this, pos.first, pos.second));
    }

    chance = step(chance);
    #ifdef DRAW
    for (int x = 0; x < WIDTH; ++x)
        for (int y = 0; y < HEIGHT; ++y)
            setColor(x, y);
    #endif

    for (int i = people.size() - 1; i >= 0; --i) {
        if (people[i]->step()) {
            delete people[i];
            people.erase(people.begin() + i);
        }
    }
}

#define PT .15f // P(turning) = PT + PR
#define PR .00f // reversing
#define PS .02f // stopping
chance_t Grid::step(const chance_t& chance, const bool& stepVisible) const {
    chance_t newChance;

    float count;
    float move[4];
    for (int x = 0; x < WIDTH; ++x) {
        for (int y = 0; y < HEIGHT; ++y) {
            if (chance[x][y].p == 0)
                continue;
            if (visible[x][y] && (!stepVisible || chance[x][y].p <= WALL)) {
                newChance[x][y] = chance[x][y];
                continue;
            }

            move[0] = !visible[x-1][y] && (!stepVisible || chance[x-1][y].p <= WALL)? PR + std::max(0.f, PT - chance[x][y].dX) : 0;
            move[1] = !visible[x][y-1] && (!stepVisible || chance[x][y-1].p <= WALL)? PR + std::max(0.f, PT - chance[x][y].dY) : 0;
            move[2] = !visible[x+1][y] && (!stepVisible || chance[x+1][y].p <= WALL)? PR + std::max(0.f, PT + chance[x][y].dX) : 0;
            move[3] = !visible[x][y+1] && (!stepVisible || chance[x][y+1].p <= WALL)? PR + std::max(0.f, PT + chance[x][y].dY) : 0;
            count = PS + move[0] + move[1] + move[2] + move[3];

            if (move[0] != 0) newChance[x-1][y] += p_t(chance[x][y].p * move[0] / count, -1, 0);
            if (move[1] != 0) newChance[x][y-1] += p_t(chance[x][y].p * move[1] / count, 0, -1);
            if (move[2] != 0) newChance[x+1][y] += p_t(chance[x][y].p * move[2] / count, 1, 0);
            if (move[3] != 0) newChance[x][y+1] += p_t(chance[x][y].p * move[3] / count, 0, 1);
            newChance[x][y] += p_t(chance[x][y].p * PS / count, 0, 0);
        }
    }
    return newChance;
}

void Grid::draw(sf::RenderWindow& window) const {
    for (int x = 0; x < WIDTH; ++x)
        for (int y = 0; y < HEIGHT; ++y)
            window.draw(shape[x][y]);
}

float Grid::getChance(const int& x, const int& y) const {
    return chance[x][y].p;
}

void Grid::addChance(const int& x, const int& y, const float& to, const float& directionX, const float& directionY) {
    chance[x][y] += p_t(to, directionX, directionY);
    setColor(x, y);
}

void Grid::setWall(const int& x, const int& y) {
    chance[x][y] = p_t(WALL, 0, 0);
    setColor(x, y);
    setVisible(x, y);
}

void Grid::setVisible(const int& x, const int& y) {
    if (x >= 0 && y >= 0 && x < WIDTH && y < HEIGHT)
        ++visible[x][y];
}

void Grid::reduceVisible(const int& x, const int& y) {
    --visible[x][y];
}

bool Grid::isVisible(const int& x, const int& y) const {
    return visible[x][y];
}

bool Grid::isTaken(const int& x, const int& y) const {
    for (const Person* p : people)
        if (p->isPos(x, y))
            return true;
    return false;
}

std::pair<int, int> Grid::getExit() const {
    return exit[rand() % exit.size()];
}

void Grid::setColor(const int& x, const int& y) {
    float p = getChance(x, y);
    if (p <= WALL)
        shape[x][y].setFillColor(sf::Color::Red);
    else if (p <= 0)
        shape[x][y].setFillColor(visible[x][y] ? sf::Color(96,96,96) : sf::Color::Black);
    else if (!visible[x][y])
        shape[x][y].setFillColor(sf::Color(0, std::min(255.f, std::sqrt(std::max(0.f, p)) * 255), std::max(0.f, std::sqrt(p - 1.f) * 255)));
    else if (p >= 1)
        shape[x][y].setFillColor(sf::Color::Green);
    else
        shape[x][y].setFillColor(sf::Color(96,96,96));
}

struct Node {
    int x, y;
    float cost;
    unsigned int d, h;

    Node(const int& x, const int& y) : x(x), y(y), cost(0), d(0), h(0) { }
    Node(const int& x, const int& y, const float& cost, const unsigned int& h, const unsigned int& d) : x(x), y(y), cost(cost), d(d), h(h) { }

    inline unsigned int getCost() const {
        return cost + d + h;
    }

    inline bool operator == (const Node b) const {
        return x == b.x && y == b.y;
    }

    inline bool operator < (const Node& b) const {
        return cost + d + h > b.cost + b.d + b.h;
    }
};
namespace std {
    template <> struct hash<Node> {
        std::size_t operator() (const Node& n) const {
            return n.x * WIDTH + n.y;
        }
    };
}

inline void checkSquare(const float& weight, const std::vector<chance_t>& prediction, std::unordered_set<Node>& explored, std::priority_queue<Node>& frontier, std::unordered_map<Node, Node>& prev,
                        const Node& parent, const int& goalX, const int& goalY, const int& x, const int& y) {
    if (prediction[parent.d][x][y].p < 1) {
        const float p = prediction[parent.d + 1][x][y].p;
        if (p > WALL && p < 1) {
            Node node(x, y, parent.cost + std::max(0.f, weight * p), std::abs(x - goalX) + std::abs(y - goalY), parent.d + 1);
            auto it = explored.find(node);
            if (it == explored.end() || it->getCost() > node.getCost()) {
                if (it != explored.end())
                    explored.erase(it);
                explored.insert(node);
                frontier.push(node);
                prev.insert({node, parent});
            }
        }
    }
}

const std::stack<int> Grid::getPath(const int& startX, const int& startY, const int& goalX, const int& goalY, const float& weight, const float& decrease) const {
    if (getChance(goalX, goalY) <= WALL)
        return std::stack<int>();

    std::priority_queue<Node> frontier;
    std::unordered_set<Node> explored;
    std::unordered_map<Node, Node> prev;
    std::vector<Node> neighbors;
    std::vector<chance_t> prediction = { chance };
    {
        Node start = Node(startX, startY);
        frontier.push(start);
        explored.insert(start);
    }
    int x, y;
    float w;

    while (!frontier.empty()) {
        Node current = frontier.top();
        frontier.pop();

        if (current.x == goalX && current.y == goalY) {
            std::stack<int> path;
            for (auto it = prev.find(current); it != prev.end(); it = prev.find(it->second)) {
                path.push(it->first.y);
                path.push(it->first.x);
            }
            return path;
        }

        x = current.x;
        y = current.y;
        w = weight - decrease * current.d;
        if (w > 0) {
            if (prediction.size() == current.d + 1)
                prediction.push_back(step(prediction[current.d], true));
        } else {
            w = 0;
            if (prediction.size() == current.d + 1)
                prediction.push_back(chance);
        }

        checkSquare(w, prediction, explored, frontier, prev, current, goalX, goalY, x + 1, y);
        checkSquare(w, prediction, explored, frontier, prev, current, goalX, goalY, x - 1, y);
        checkSquare(w, prediction, explored, frontier, prev, current, goalX, goalY, x, y + 1);
        checkSquare(w, prediction, explored, frontier, prev, current, goalX, goalY, x, y - 1);
        checkSquare(w, prediction, explored, frontier, prev, current, goalX, goalY, x,     y);
    }
    return std::stack<int>();
}
