#ifndef GRID_H_INCLUDED
#define GRID_H_INCLUDED

#include <array>
#include <stack>

#include <SFML/Graphics/RectangleShape.hpp>
namespace sf {
    class Image;
    class RenderWindow;
}

class Person;

#define WIDTH 58
#define HEIGHT 53
#define SIZE 16
#define WALL -16

struct p_t {
    float p, dX, dY;
    p_t(const float& p, const float& dX, const float& dY) : p(p), dX(dX), dY(dY) {}
    p_t() : p(0) {}
    void operator+=(const p_t& a) {
        if (p == 0) {
            p = a.p;
            dX = a.dX;
            dY = a.dY;
        } else if (a.p != 0) {
            float ap = std::abs(p), aap = std::abs(a.p);
            dX = (dX * ap + a.dX * aap) / (ap + aap);
            dY = (dY * ap + a.dY * aap) / (ap + aap);
            p = p + a.p;
        }
    }
};
typedef std::array<std::array<p_t, HEIGHT>, WIDTH> chance_t;

class Grid {
public:
    Grid(const sf::Image& map, int& robotX, int& robotY);
    void step();
    void draw(sf::RenderWindow& window) const;
    float getChance(const int& x, const int& y) const;
    void addChance(const int& x, const int& y, const float& to, const float& directionX, const float& directionY);
    void setWall(const int& x, const int& y);
    void setVisible(const int& x, const int& y);
    void reduceVisible(const int& x, const int& y);
    bool isVisible(const int& x, const int& y) const;
    bool isTaken(const int& x, const int& y) const;
    std::pair<int, int> getExit() const;
    const std::stack<int> getPath(const int& startX, const int& startY, const int& goalX, const int& goalY, const float& weight = 0.f, const float& decrease = 0.f) const;
private:
    chance_t chance;
    unsigned short visible[WIDTH][HEIGHT];
    sf::RectangleShape shape[WIDTH][HEIGHT];
    std::vector<std::pair<int, int>> entrance;
    std::vector<std::pair<int, int>> exit;
    std::vector<Person*> people;

    chance_t step(const chance_t& chance, const bool& stepVisible = false) const;
    void setColor(const int& x, const int& y);
    void makeCamera(const int& x, const int& y);
};

#endif // GRID_H_INCLUDED
