#include "person.h"

#include <iostream>

#include "grid.h"


std::poisson_distribution<unsigned int> Person::poisson(20);

Person::Person(Grid& grid, const int& x, const int& y) : grid(grid), x(x), y(y), directionX(0), directionY(0) {
    grid.addChance(x, y, 1, directionX, directionY);

    target.push(grid.getExit());
    std::default_random_engine generator(rand());
    unsigned int i = poisson(generator);
    for (; i > 0; --i)  // On average, people stay for ~500 time units
        target.push(std::make_pair(rand() % WIDTH, rand() % HEIGHT));
}

Person::~Person() {
    grid.addChance(x, y, -1, directionX, directionY);
}

bool Person::step() {
    if (path.empty()) {
        if (target.empty())
            return true;
        std::pair<int, int> goal = target.top();
        target.pop();
        path = grid.getPath(x, y, goal.first, goal.second);
    } else {
        int nx, ny;
        nx = path.top();
        path.pop();
        ny = path.top();
        path.pop();

        if (grid.isVisible(x, y) || grid.isVisible(nx, ny)) {
            grid.addChance(x, y, -1, -directionX, -directionY);
            grid.addChance(nx, ny, 1, nx - x, ny - y);
        }
        directionX = nx - x;
        directionY = ny - y;
        x = nx;
        y = ny;
    }
    return false;
}
