#include "robot.h"

#include <iostream>

#include <SFML/Graphics/RenderWindow.hpp>

#include "grid.h"


sf::Font Robot::font;

Robot::Robot(Grid& grid, const int& x, const int& y, const float& weight) : grid(grid), x(x), y(y), targetX(x), targetY(y), direction(0), weight(weight) {
    setVisible();
    shape.setSize(sf::Vector2f(SIZE, SIZE));
    shape.setFillColor(weight == 0 ? sf::Color::Cyan : sf::Color::Blue);
    shape.setPosition(x * SIZE, y * SIZE);
    font.loadFromFile("sansation.ttf");
    text.setFont(font);
    if (weight != 0)
        text.setPosition(sf::Vector2f(WIDTH * SIZE - 64, 0));
    text.setString("0");
}

void Robot::step() {
    int nx, ny;
    while (!vision.empty()) {
        nx = vision.top();
        vision.pop();
        ny = vision.top();
        vision.pop();
        grid.reduceVisible(nx, ny);
    }

    if (targetX != x || targetY != y) {
        if (grid.isTaken(x, y))
            text.setString(std::to_string(++count));

        path = grid.getPath(x, y, targetX, targetY, weight);
        if (!path.empty()) {
            nx = path.top();
            path.pop();
            ny = path.top();
            path.pop();
            if (x != nx) {
                direction |= 1;
                if (x > nx) direction &= 1;
                else        direction |= 2;
            }
            if (y != ny) {
                direction &= 2;
                if (y > ny) direction &= 1;
                else        direction |= 2;
            }
            x = nx;
            y = ny;
            shape.setPosition(x * SIZE, y * SIZE);
        }
        #ifndef DRAW
        else {
            targetX = rand() % WIDTH;
            targetY = rand() % HEIGHT;
            if (++targets > 50)
                throw count;
        }
    } else {
        targetX = rand() % WIDTH;
        targetY = rand() % HEIGHT;
        if (++targets > 50)
            throw count;
        #endif
    }

    setVisible();
}

void Robot::draw(sf::RenderWindow& window) const {
    window.draw(text);
    window.draw(shape);
}

void Robot::setTarget(const int& x, const int& y) {
    if (x < WIDTH && y < HEIGHT) {
        targetX = x;
        targetY = y;
    }
}

void Robot::setVisible() {
    int ch;
    if (direction & 1) {
        for (int j = y - 1; j <= y + 1; ++j) {
            for (int i = direction & 2 ? x + 1 : x - 1; i >= 0 && i < WIDTH; direction & 2 ? ++i : --i) {
                grid.setVisible(i, j);
                vision.push(j);
                vision.push(i);
                ch = grid.getChance(i, j);
                if (ch <= WALL || ch >= 1)
                    break;
            }
        }
    } else {
        for (int i = x - 1; i <= x + 1; ++i) {
            for (int j = direction & 2 ? y + 1 : y - 1; j >= 0 && j < HEIGHT; direction & 2 ? ++j : --j) {
                grid.setVisible(i, j);
                vision.push(j);
                vision.push(i);
                ch = grid.getChance(i, j);
                if (ch <= WALL || ch >= 1)
                    break;
            }
        }
    }
}
